/**
 * Alternating Bit Protocol
 *
 * 'Buggy Code'
 *
 * Explanation: both the Sender and the Receiver might happen to wait
 *     for a message to be received, while neither of the two is sending
 *     anything. Thus, the processes incur into deadlock.
 *
 * Notice that writing:
 * 1A do
 * 2A     :: channel01?message(...);
 * 3A     :: channel02?message(...);
 * 4A od;
 * and writing:
 * 1B do
 * 2B     :: true -> channel01?message(...);
 * 3B     :: true -> channel02?message(...);
 * 4B od;
 * is not semantically equivalent!
 *
 * If a process executes the code in lines 1A-4A, it will block
 * at line 1A until either one of the two channels provides some message
 * to be read.
 *
 * If a process executes the code in lines 1B-4B, it will either block
 * at line 2B waiting for a message coming from channel01, or block at line
 * 3B waiting for a message coming from channel02.
 *
 * Learning this difference is part of coming up with a solution for 
 * this exercise (see: fixed_unreliable_altbit_protocol.pml).
 */

mtype = { MESSAGE, ACK };

chan sender2receiver = [2] of { mtype, bit, int};
chan receiver2sender = [2] of { mtype, bit, int};

inline unreliable_send(channel, type, tag, seqno) {
    bool loss = false;
    bool duplicate = true;
    if
        :: channel!type(tag, seqno);
            if
                :: channel!type(tag, seqno);
                :: duplicate = false;
            fi
        :: loss = true;
    fi
}

active proctype Sender () {
    bit in_bit, out_bit;
    int seq_no;

    do
        :: unreliable_send(sender2receiver, MESSAGE, out_bit, seq_no) ->
            receiver2sender?ACK(in_bit, 0);
            if
                :: in_bit == out_bit ->
                    out_bit = 1 - out_bit;
                    seq_no++;
                :: else ->
                    skip;
            fi;
    od;
}

active proctype Receiver () {
    bit in_bit, old_bit;
    int seq_no;

    do
        :: sender2receiver?MESSAGE(in_bit, seq_no) ->
            if
                :: in_bit != old_bit ->
                    printf("received: %d\n", seq_no);
                    old_bit = in_bit;
                :: else ->
                    skip;
            fi;
            unreliable_send(receiver2sender, ACK, in_bit, 0);
    od;
}
